from abc import ABCMeta, abstractmethod
from dataclasses import dataclass
from typing import Generic, List, NoReturn, TypeVar

from ziopy.either import Either
from ziopy.hlist import HCons, HList, HNil

A = TypeVar("A")
E = TypeVar("E")
H = TypeVar("H")
T = TypeVar("T", bound=HList)
U = TypeVar("U", bound="Validation")


class Validation(Generic[E, T], metaclass=ABCMeta):
    @abstractmethod
    def validate(self) -> Either[List[E], T]:
        pass


@dataclass(frozen=True)
class VNil(Validation[NoReturn, HNil]):
    def prepend(self, h: Either[E, H]) -> "VCons[E, H, VNil, HCons[H, HNil]]":
        return VCons(h, self)

    def __rpow__(self, h: Either[E, H]) -> "VCons[E, H, VNil, HCons[H, HNil]]":
        return self.prepend(h)

    def validate(self) -> Either[List[E], HNil]:
        return Either.right(HNil())


@dataclass(frozen=True)
class VCons(Generic[E, H, U, T], Validation[E, T]):
    head: Either[E, H]
    tail: U

    def prepend(self, h: Either[E, A]) -> "VCons[E, A, VCons[E, H, U, T], HCons[A, T]]":
        return VCons(h, self)

    def __rpow__(self, h: Either[E, A]) -> "VCons[E, A, VCons[E, H, U, T], HCons[A, T]]":
        return self.prepend(h)

    def validate(self) -> Either[List[E], T]:
        tail_v = self.tail.validate()

        return (
            self.head
            .map_left(
                lambda e: tail_v.fold(lambda errs: [e] + errs, lambda _: [e])
            )
            .flat_map(lambda t1: tail_v.map(lambda rs: (t1 ** rs)))
        )



from dataclasses import dataclass

import ziopy.services.console as console
import ziopy.services.system as system


@dataclass(frozen=True)
class ConsoleEnvironment:
    console: console.Console


@dataclass(frozen=True)
class SystemEnvironment:
    system: system.System


@dataclass(frozen=True)
class ConsoleSystemEnvironment(ConsoleEnvironment, SystemEnvironment):
    pass



import sys
from abc import ABCMeta, abstractmethod
from typing import List, NoReturn, Optional
from typing_extensions import Protocol

import ziopy.services.mock_effects.system as system_effect
from ziopy.zio import ZIO, Environment


class System(metaclass=ABCMeta):
    @abstractmethod
    def exit(self, exit_code: Optional[int] = None) -> ZIO[object, NoReturn, NoReturn]:
        pass  # pragma: nocover


class LiveSystem(System):
    def exit(self, exit_code: Optional[int] = None) -> ZIO[object, NoReturn, NoReturn]:
        return ZIO.effect_total(lambda: sys.exit(exit_code))  # type: ignore


class MockSystem(System):
    def __init__(self) -> None:
        self._effects: List[system_effect.Exit] = []

    def exit(self, exit_code: Optional[int] = None) -> ZIO[object, NoReturn, NoReturn]:
        self._effects.append(system_effect.Exit(exit_code))
        return ZIO.fail(SystemExit(exit_code)).or_die()

    @property
    def effects(self) -> List[system_effect.Exit]:
        return self._effects


class HasSystem(Protocol):
    @property
    def system(self) -> System:
        pass  # pragma: nocover


def exit(exit_code: Optional[int] = None) -> ZIO[HasSystem, NoReturn, NoReturn]:
    return Environment[HasSystem]().flat_map(lambda env: env.system.exit(exit_code))



import functools
from abc import ABCMeta
from dataclasses import dataclass
from typing import Callable, Generic, TypeVar, overload

from typing_extensions import TypeAlias

A = TypeVar("A")
R = TypeVar("R")
H = TypeVar("H", covariant=True)

T = TypeVar("T", covariant=True, bound="HList")


class HList(metaclass=ABCMeta):
    def prepend(self: T, head: A) -> "HCons[A, T]":
        return HCons(head, self)

    def __rpow__(self: T, head: A) -> "HCons[A, T]":
        return self.prepend(head)


@dataclass(frozen=True)
class HNil(HList):
    def apply(self, f: Callable[[], R]) -> R:
        return f()

    def apply_0(self, f: Callable[[], R]) -> R:
        return f()


@dataclass(frozen=True)
class HCons(Generic[H, T], HList):
    head: H
    tail: T


T1 = TypeVar("T1")
T2 = TypeVar("T2")
T3 = TypeVar("T3")
T4 = TypeVar("T4")
T5 = TypeVar("T5")
T6 = TypeVar("T6")
T7 = TypeVar("T7")
T8 = TypeVar("T8")
T9 = TypeVar("T9")
T10 = TypeVar("T10")
T11 = TypeVar("T11")
T12 = TypeVar("T12")
T13 = TypeVar("T13")
T14 = TypeVar("T14")
T15 = TypeVar("T15")


# TODO: If mypy's ParamSpec actually worked as expected, lots of this boilerplate
# would go away. In the meantime, behold the horror.
HList1: TypeAlias = HCons[T1, HNil]
HList2: TypeAlias = HCons[T1, HList1[T2]]
HList3: TypeAlias = HCons[T1, HList2[T2, T3]]
HList4: TypeAlias = HCons[T1, HList3[T2, T3, T4]]
HList5: TypeAlias = HCons[T1, HList4[T2, T3, T4, T5]]
HList6: TypeAlias = HCons[T1, HList5[T2, T3, T4, T5, T6]]
HList7: TypeAlias = HCons[T1, HList6[T2, T3, T4, T5, T6, T7]]
HList8: TypeAlias = HCons[T1, HList7[T2, T3, T4, T5, T6, T7, T8]]
HList9: TypeAlias = HCons[T1, HList8[T2, T3, T4, T5, T6, T7, T8, T9]]
HList10: TypeAlias = HCons[T1, HList9[T2, T3, T4, T5, T6, T7, T8, T9, T10]]
HList11: TypeAlias = HCons[T1, HList10[T2, T3, T4, T5, T6, T7, T8, T9, T10, T11]]


@overload
def hlisted(f: Callable[[], R]) -> Callable[[HNil], R]: ...
@overload
def hlisted(f: Callable[[T1], R]) -> Callable[[HList1[T1]], R]: ...
@overload
def hlisted(f: Callable[[T1, T2], R]) -> Callable[[HList2[T1, T2]], R]: ...
@overload
def hlisted(f: Callable[[T1, T2, T3], R]) -> Callable[[HList3[T1, T2, T3]], R]: ...
@overload
def hlisted(f: Callable[[T1, T2, T3, T4], R]) -> Callable[[HList4[T1, T2, T3, T4]], R]: ...
@overload
def hlisted(f: Callable[[T1, T2, T3, T4, T5], R]) -> Callable[[HList5[T1, T2, T3, T4, T5]], R]: ...
@overload
def hlisted(f: Callable[[T1, T2, T3, T4, T5, T6], R]) -> Callable[[HList6[T1, T2, T3, T4, T5, T6]], R]: ...
@overload
def hlisted(f: Callable[[T1, T2, T3, T4, T5, T6, T7], R]) -> Callable[[HList7[T1, T2, T3, T4, T5, T6, T7]], R]: ...
@overload
def hlisted(f: Callable[[T1, T2, T3, T4, T5, T6, T7, T8], R]) -> Callable[[HList8[T1, T2, T3, T4, T5, T6, T7, T8]], R]: ...
@overload
def hlisted(f: Callable[[T1, T2, T3, T4, T5, T6, T7, T8, T9], R]) -> Callable[[HList9[T1, T2, T3, T4, T5, T6, T7, T8, T9]], R]: ...
@overload
def hlisted(f: Callable[[T1, T2, T3, T4, T5, T6, T7, T8, T9, T10], R]) -> Callable[[HList10[T1, T2, T3, T4, T5, T6, T7, T8, T9, T10]], R]: ...


def hlisted(f: Callable[..., R]) -> Callable[..., R]:
    def g(hlist: HList) -> R:
        if isinstance(hlist, HNil):
            return f()
        elif isinstance(hlist, HCons):
            # Partially apply f with Head=T1 to get a function
            # `T2, T3, T4, ... -> R`
            f_partial = functools.partial(f, hlist.head)

            # now hlist the partially applied function
            f_partial_hlisted = hlisted(f_partial)

            # Apply it to the tail
            return f_partial_hlisted(hlist.tail)
        else:
            raise TypeError

    return g



import functools
from dataclasses import dataclass
from typing import Callable, Generic, NoReturn, Tuple, Type, TypeVar, Union

from typing_extensions import Concatenate, ParamSpec

from ziopy.either import Either, Left, Right
from ziopy.util import RaiseLeft, raise_exception

"""
Heavily inspired by:
https://github.com/jdegoes/functional-effects/blob/master/src/main/scala/net/degoes/zio/00-intro.scala
"""


R = TypeVar('R', contravariant=True)
E = TypeVar('E', covariant=True)
A = TypeVar('A', covariant=True)
B = TypeVar('B')

G = TypeVar('G', bound=BaseException, covariant=True)

RR = TypeVar('RR')
EE = TypeVar('EE')
AA = TypeVar('AA')
BB = TypeVar('BB')

E2 = TypeVar('E2')
A2 = TypeVar('A2')

T = TypeVar('T')
Thunk = Callable[[], T]

P = ParamSpec("P")
F = TypeVar('F', bound=Callable)

X = TypeVar('X', bound=BaseException)


@dataclass(frozen=True)
class TypeMatchException(Generic[AA], Exception):
    value: AA


class ZIO(Generic[R, E, A]):
    def __init__(self, run: Callable[[R], Either[E, A]]):
        self._run = run

    @staticmethod
    def succeed(a: AA) -> "ZIO[object, NoReturn, AA]":
        return ZIO(lambda _: Right(a))

    @staticmethod
    def fail(e: EE) -> "ZIO[object, EE, NoReturn]":
        return ZIO(lambda _: Left(e))

    @staticmethod
    def from_either(e: Either[EE, AA]) -> "ZIO[object, EE, AA]":
        return ZIO(lambda _: e)

    @staticmethod
    def effect(side_effect: Thunk[AA]) -> "ZIO[object, Exception, AA]":
        a = ZIO[object, NoReturn, AA](lambda _: Either.right(side_effect()))
        return a.catch(Exception)

    @staticmethod
    def effect_catch(side_effect: Thunk[AA], exception_type: Type[X]) -> "ZIO[object, X, AA]":
        a = ZIO[object, NoReturn, AA](lambda _: Either.right(side_effect()))
        return a.catch(exception_type)

    @staticmethod
    def access(f: Callable[[R], AA]) -> "ZIO[R, NoReturn, AA]":
        return ZIO(lambda r: Right(f(r)))

    @staticmethod
    def access_m(f: Callable[[RR], "ZIO[RR, EE, AA]"]) -> "ZIO[RR, EE, AA]":
        return ZIO(lambda r: f(r)._run(r))

    def provide(self, r: R) -> "ZIO[object, E, A]":
        return ZIO(lambda _: self._run(r))

    @staticmethod
    def effect_total(side_effect: Thunk[AA]) -> "ZIO[object, NoReturn, AA]":
        return ZIO(lambda _: Right(side_effect()))

    def catch(
        self: "ZIO[R, E, AA]",
        exc: Type[X]
    ) -> "ZIO[R, Union[E, X], AA]":
        def _f(r: R) -> Either[Union[E, X], AA]:
            try:
                return self._run(r)
            except exc as e:
                return Either.left(e)
        return ZIO(_f)

    def map(self, f: Callable[[A], B]) -> "ZIO[R, E, B]":
        return ZIO(lambda r: self._run(r).map(f))

    def map_error(self: "ZIO[RR, EE, AA]", f: Callable[[EE], E2]) -> "ZIO[RR, E2, AA]":
        return ZIO(lambda r: self._run(r).map_left(f))

    def flat_map(
        self: "ZIO[RR, E, AA]",
        f: Callable[[AA], "ZIO[RR, EE, B]"]
    ) -> "ZIO[RR, Union[E, EE], B]":
        return ZIO(lambda r: self._run(r).flat_map(lambda a: f(a)._run(r)))

    def flatten(
        self: "ZIO[R, E, ZIO[R, EE, AA]]"
    ) -> "ZIO[R, Union[E, EE], AA]":
        return self.flat_map(lambda x: x)

    def __lshift__(self: "ZIO[RR, EE, AA]", other: "ZIO[RR, EE, B]") -> "ZIO[RR, EE, B]":
        return self.flat_map(lambda _: other)

    def zip(
        self: "ZIO[RR, E, AA]",
        that: "ZIO[RR, EE, B]"
    ) -> "ZIO[RR, Union[E, EE], Tuple[AA, B]]":
        return self.flat_map(lambda a: that.map(lambda b: (a, b)))

    def either(self) -> "ZIO[R, NoReturn, Either[E, A]]":
        return ZIO(lambda r: Right(self._run(r)))

    def absolve(self: "ZIO[R, E, Either[EE, AA]]") -> "ZIO[R, Union[E, EE], AA]":
        return self.flat_map(ZIO.from_either)

    def or_die(self: "ZIO[R, X, AA]") -> "ZIO[R, NoReturn, AA]":
        return ZIO(lambda r: self._run(r).fold(raise_exception, lambda a: Right(a)))

    def require(
        self: "ZIO[R, E, AA]",
        predicate: Callable[[AA], bool],
        to_error: Callable[[AA], EE]
    ) -> "ZIO[R, Union[E, EE], AA]":
        return ZIO(lambda r: self._run(r).require(predicate, to_error))

    def asserting(
        self: "ZIO[R, E, AA]",
        predicate: Callable[[AA], bool],
        to_error: Callable[[AA], X]
    ) -> "ZIO[R, E, AA]":
        return ZIO(lambda r: self._run(r).asserting(predicate, to_error))

    def or_else(
        self: "ZIO[R, EE, AA]",
        other: "ZIO[R, E2, A2]"
    ) -> "ZIO[R, Union[EE, E2], Union[AA, A2]]":
        return ZIO(
            lambda r: self._run(r).fold(
                lambda e: other._run(r),
                lambda a: Right(a)
            )
        )

    def swap(self: "ZIO[R, EE, AA]") -> "ZIO[R, AA, EE]":
        return ZIO(lambda r: self._run(r).swap())

    def match_types(self: "ZIO[R, E, AA]") -> "ZIO[R, E, NoReturn]":
        def _f(arg: AA) -> NoReturn:
            raise TypeMatchException(arg)
        return self.map(_f)

    def at_type(
        self: "ZIO[RR, EE, AA]",
        target_type: Type[A2],
        operation: "ZIO[A2, E2, BB]"
    ) -> "ZIO[RR, Union[EE, E2], Union[AA, BB]]":
        def _recover(arg: Union[EE, TypeMatchException]) -> ZIO[object, Union[EE, E2], BB]:
            if not isinstance(arg, TypeMatchException):
                return ZIO.fail(arg)
            if isinstance(arg.value, target_type):
                return operation.provide(arg.value)
            else:
                raise arg from arg

        return (
            self
            .catch(TypeMatchException)
            .swap()
            .either()
            .flat_map(lambda e: e.fold(ZIO.succeed, _recover))
        )


class Environment(Generic[RR], ZIO[RR, NoReturn, RR]):
    def __init__(self) -> None:
        self._run = lambda r: Right(r)


def unsafe_run(io: ZIO[object, X, AA]) -> AA:
    return io._run(None).fold(raise_exception, lambda a: a)


class ZIOMonad(Generic[R, EE]):
    def __init__(self, environment: R) -> None:
        self._environment = environment

    def __lshift__(self, arg: ZIO[R, EE, BB]) -> BB:
        return arg._run(self._environment).fold(
            lambda e: raise_exception(RaiseLeft(e)),
            lambda a: a
        )


def monadic_method(
    func: Callable[Concatenate[AA, ZIOMonad[R, E], P], ZIO[R, E, A]]
) -> Callable[Concatenate[AA, P], ZIO[R, E, A]]:
    @functools.wraps(func)
    def _wrapper(self_arg: AA, *args: P.args, **kwargs: P.kwargs) -> ZIO[R, E, A]:
        def _catch_left(environment: R) -> ZIO[R, E, A]:
            try:
                return func(self_arg, ZIOMonad(environment), *args, **kwargs)
            except RaiseLeft as raise_left:
                # NOTE: WJH (12/20/20) mypy can't prove that the generic type
                #       of the _RaiseLeft instance here is `E`, so we have to
                #       use `type: ignore`.
                return ZIO.fail(raise_left.value)  # type: ignore

        return Environment[R]().flat_map(_catch_left)
    return _wrapper


def monadic(
    func: Callable[Concatenate[ZIOMonad[R, E], P], ZIO[R, E, A]]
) -> Callable[P, ZIO[R, E, A]]:
    @functools.wraps(func)
    def _wrapper(*args: P.args, **kwargs: P.kwargs) -> ZIO[R, E, A]:
        def _catch_left(environment: R) -> ZIO[R, E, A]:
            try:
                return func(ZIOMonad(environment), *args, **kwargs)
            except RaiseLeft as raise_left:
                # NOTE: WJH (12/20/20) mypy can't prove that the generic type
                #       of the _RaiseLeft instance here is `E`, so we have to
                #       use `type: ignore`.
                return ZIO.fail(raise_left.value)  # type: ignore

        return Environment[R]().flat_map(_catch_left)
    return _wrapper


class ZIOArrow(Generic[P, R, E, A]):
    @staticmethod
    def from_zio(zio: ZIO[R, E, Callable[P, A]]) -> "ZIOArrow[P, R, E, A]":
        return ZIOArrow(lambda *args, **kwargs: zio.map(lambda f: f(*args, **kwargs)))

    @staticmethod
    def from_callable(f: Callable[P, A]) -> "ZIOArrow[P, object, NoReturn, A]":
        return ZIOArrow(lambda *args, **kwargs: ZIO.succeed(f(*args, **kwargs)))

    def __init__(self, f: Callable[P, ZIO[R, E, A]]) -> None:
        self._f = f

    def to_callable(self) -> "Callable[P, ZIO[R, E, A]]":
        return self._f

    def __call__(self, *args: P.args, **kwargs: P.kwargs) -> "ZIO[R, E, A]":
        return self._f(*args, **kwargs)

    def provide(self, r: R) -> "ZIOArrow[P, object, E, A]":
        return ZIOArrow(lambda *args, **kwargs: self._f(*args, **kwargs).provide(r))

    def catch(
        self: "ZIOArrow[P, R, E, AA]",
        exc: Type[X]
    ) -> "ZIOArrow[P, R, Union[E, X], AA]":
        def _g(*args: P.args, **kwargs: P.kwargs) -> ZIO[R, Union[E, X], AA]:
            try:
                return self._f(*args, **kwargs)
            except exc as e:
                return ZIO.fail(e)
        return ZIOArrow(_g)

    def map(self, f: Callable[[A], B]) -> "ZIOArrow[P, R, E, B]":
        return ZIOArrow(lambda *args, **kwargs: self._f(*args, **kwargs).map(f))

    def map_error(
        self: "ZIOArrow[P, RR, EE, AA]", f: Callable[[EE], E2]
    ) -> "ZIOArrow[P, RR, E2, AA]":
        return ZIOArrow(lambda *args, **kwargs: self._f(*args, **kwargs).map_error(f))

    def flat_map(
        self: "ZIOArrow[P, RR, E, AA]",
        f: Callable[[AA], "ZIO[RR, EE, B]"]
    ) -> "ZIOArrow[P, RR, Union[E, EE], B]":
        return ZIOArrow(lambda *args, **kwargs: self._f(*args, **kwargs).flat_map(f))

    def flatten(
        self: "ZIOArrow[P, R, E, ZIO[R, EE, AA]]"
    ) -> "ZIOArrow[P, R, Union[E, EE], AA]":
        return self.flat_map(lambda x: x)

    def __lshift__(
        self: "ZIOArrow[P, RR, EE, AA]", other: "ZIO[RR, EE, B]"
    ) -> "ZIOArrow[P, RR, EE, B]":
        return self.flat_map(lambda _: other)

    def zip(
        self: "ZIOArrow[P, RR, E, AA]",
        that: "ZIO[RR, EE, B]"
    ) -> "ZIOArrow[P, RR, Union[E, EE], Tuple[AA, B]]":
        return self.flat_map(lambda a: that.map(lambda b: (a, b)))

    def either(self) -> "ZIOArrow[P, R, NoReturn, Either[E, A]]":
        return ZIOArrow(lambda *args, **kwargs: self._f(*args, **kwargs).either())

    def absolve(
        self: "ZIOArrow[P, R, E, Either[EE, AA]]"
    ) -> "ZIOArrow[P, R, Union[E, EE], AA]":
        return self.flat_map(ZIO.from_either)

    def or_die(self: "ZIOArrow[P, R, X, AA]") -> "ZIOArrow[P, R, NoReturn, AA]":
        return ZIOArrow(lambda *args, **kwargs: self._f(*args, **kwargs).or_die())

    def require(
        self: "ZIOArrow[P, R, E, AA]",
        predicate: Callable[[AA], bool],
        to_error: Callable[[AA], EE]
    ) -> "ZIOArrow[P, R, Union[E, EE], AA]":
        return ZIOArrow(
            lambda *args, **kwargs: self._f(*args, **kwargs).require(predicate, to_error)
        )

    def asserting(
        self: "ZIOArrow[P, R, E, AA]",
        predicate: Callable[[AA], bool],
        to_error: Callable[[AA], X]
    ) -> "ZIOArrow[P, R, E, AA]":
        return ZIOArrow(
            lambda *args, **kwargs: self._f(*args, **kwargs).asserting(predicate, to_error)
        )

    def or_else(
        self: "ZIOArrow[P, R, EE, AA]",
        other: "ZIO[R, E2, A2]"
    ) -> "ZIOArrow[P, R, Union[EE, E2], Union[AA, A2]]":
        return ZIOArrow(lambda *args, **kwargs: self._f(*args, **kwargs).or_else(other))

    def swap(self: "ZIOArrow[P, R, EE, AA]") -> "ZIOArrow[P, R, AA, EE]":
        return ZIOArrow(lambda *args, **kwargs: self._f(*args, **kwargs).swap())

    def match_types(self: "ZIOArrow[P, R, E, AA]") -> "ZIOArrow[P, R, E, NoReturn]":
        return ZIOArrow(lambda *args, **kwargs: self._f(*args, **kwargs).match_types())

    def at_type(
        self: "ZIOArrow[P, RR, EE, AA]",
        target_type: Type[A2],
        operation: "ZIO[A2, E2, BB]"
    ) -> "ZIOArrow[P, RR, Union[EE, E2], Union[AA, BB]]":
        return ZIOArrow(
            lambda *args, **kwargs: self._f(*args, **kwargs).at_type(target_type, operation)
        )



from dataclasses import dataclass
from typing import Generic, NoReturn, TypeVar


E = TypeVar("E", covariant=True)


def raise_exception(x: BaseException) -> NoReturn:
    raise x


@dataclass(frozen=True)
class RaiseLeft(Generic[E], Exception):
    value: E



import builtins
from abc import ABCMeta, abstractmethod
from typing import Callable, List, NoReturn, Optional, TypeVar, Union

import ziopy.services.mock_effects.console as console_effect
from typing_extensions import Literal, Protocol
from ziopy.either import Either, EitherArrow, Right
from ziopy.services.system import HasSystem, System
from ziopy.zio import ZIO, Environment, ZIOArrow, ZIOMonad, monadic_method

A = TypeVar('A')
E = TypeVar('E')


class Console(metaclass=ABCMeta):
    @abstractmethod
    def print(self, line: str) -> ZIO[object, NoReturn, None]:
        pass  # pragma: nocover

    @abstractmethod
    def input(
        self,
        prompt: Optional[str] = None
    ) -> ZIO[object, Union[EOFError, KeyboardInterrupt], str]:
        pass  # pragma: nocover

    @monadic_method
    def get_input_from_console(
        self,
        do: ZIOMonad[System, NoReturn],
        prompt: str,
        parse_value: Callable[[str], Either[E, A]],
        default_value: Optional[A]
    ) -> ZIO[System, NoReturn, A]:
        while True:
            keyboard_input = do << (
                self.input(prompt)
                .either()
                .map(lambda e: e.to_union())
            )

            if isinstance(keyboard_input, (EOFError, KeyboardInterrupt)):
                do << self.print("")
                system = do << Environment()
                return system.exit()

            if keyboard_input == '' and default_value is not None:
                return ZIO.succeed(default_value)

            parse_result = parse_value(keyboard_input)
            if isinstance(parse_result, Right):
                return ZIO.succeed(parse_result.value)

    @monadic_method
    def ask(
        self,
        do: ZIOMonad[System, NoReturn],
        prompt: str,
        default: Literal['y', 'n']
    ) -> ZIO[System, NoReturn, bool]:
        default_str = 'Y/n' if default == 'y' else 'y/N'
        choice = do << self.get_input_from_console(
            prompt=f"{prompt} [{default_str}]: ",
            parse_value=(
                EitherArrow.from_callable(str)
                .map(str.lower)
                .require(lambda s: s in {'y', 'n'}, lambda s: s)
            ),
            default_value=default
        )
        return ZIO.succeed(choice == 'y')


class LiveConsole(Console):
    def print(self, line: str) -> ZIO[object, NoReturn, None]:
        return ZIO.effect_total(lambda: builtins.print(line))

    def input(
        self,
        prompt: Optional[str] = None
    ) -> ZIO[object, Union[EOFError, KeyboardInterrupt], str]:
        return (
            ZIO.effect_catch(
                lambda: builtins.input(prompt) if prompt is not None else builtins.input(),
                EOFError
            )
            .catch(KeyboardInterrupt)
        )


class MockConsole(Console):
    def __init__(self, user_input: Optional[List[Union[EOFError, KeyboardInterrupt, str]]] = None) -> None:
        self._effects: List[Union[console_effect.Print, console_effect.Input]] = []
        if user_input is None:
            user_input = []
        self._user_input = user_input

    def print(self, line: str) -> ZIO[object, NoReturn, None]:
        self._effects.append(console_effect.Print(line))
        return ZIO.succeed(None)

    def input(
        self,
        prompt: Optional[str] = None
    ) -> ZIO[object, Union[EOFError, KeyboardInterrupt], str]:
        user_input = self._user_input.pop(0)
        self._effects.append(console_effect.Input(prompt, user_input))
        if isinstance(user_input, str):
            return ZIO.succeed(user_input)
        else:
            result: Union[EOFError, KeyboardInterrupt] = user_input
            return ZIO.fail(result)

    @property
    def effects(self) -> List[Union[console_effect.Print, console_effect.Input]]:
        return self._effects

    @property
    def user_input(self) -> List[Union[EOFError, KeyboardInterrupt, str]]:
        return self._user_input


class HasConsole(Protocol):
    @property
    def console(self) -> Console:
        pass  # pragma: nocover


print = (
    ZIOArrow.from_zio(
        Environment[HasConsole]()
        .map(lambda env: env.console.print)
    )
    .flatten()
    .to_callable()
)

input = (
    ZIOArrow.from_zio(
        Environment[HasConsole]()
        .map(lambda env: env.console.input)
    )
    .flatten()
    .to_callable()
)


class HasConsoleSystem(HasConsole, HasSystem, Protocol):
    pass


def get_input_from_console(
    prompt: str,
    parse_value: Callable[[str], Either[E, A]],
    default_value: Optional[A]
) -> ZIO[HasConsoleSystem, NoReturn, A]:
    return (
        Environment[HasConsoleSystem]()
        .flat_map(
            lambda env: env.console.get_input_from_console(
                prompt, parse_value, default_value
            ).provide(env.system)
        )
    )


def ask(prompt: str, default: Literal['y', 'n']) -> ZIO[HasConsoleSystem, NoReturn, bool]:
    return Environment[HasConsoleSystem]().flat_map(
        lambda env: env.console.ask(prompt, default).provide(env.system)
    )



import functools
from abc import ABCMeta
from dataclasses import dataclass
from typing import (Any, Callable, Generic, NoReturn, Optional, Type, TypeVar,
                    Union)

from typing_extensions import Concatenate, Final, ParamSpec, Protocol

from ziopy.util import RaiseLeft, raise_exception

A = TypeVar('A', covariant=True)
A1 = TypeVar('A1')
A2 = TypeVar('A2')
AA = TypeVar('AA')
A_con = TypeVar("A_con", contravariant=True)

B = TypeVar('B', covariant=True)
B1 = TypeVar('B1')
B2 = TypeVar('B2')
BB = TypeVar('BB')

C = TypeVar('C')
C1 = TypeVar('C1')
C2 = TypeVar('C2')

E = TypeVar('E', covariant=True)
E1 = TypeVar('E1')
E2 = TypeVar('E2')

X = TypeVar('X', bound=BaseException)
P = ParamSpec("P")
PP = ParamSpec("PP")


@dataclass(frozen=True)
class EitherException(Generic[A], Exception):
    value: A


@dataclass(frozen=True)
class TypeMatchException(Generic[AA], Exception):
    value: Optional[AA]


class GetItem(Protocol[A_con, B]):
    def __getitem__(self, item: A_con) -> B:
        pass


class Either(Generic[A, B], metaclass=ABCMeta):
    """Right-biased disjunction"""

    @staticmethod
    def left(a: AA) -> "Either[AA, NoReturn]":
        return Left(a)

    @staticmethod
    def right(b: BB) -> "Either[NoReturn, BB]":
        return Right(b)

    @staticmethod
    def from_union(
        value: Union[A, B],
        left_type: Type[A],
        right_type: Type[B]
    ) -> "Either[A, B]":
        if isinstance(value, left_type):
            return Either.left(value)
        elif isinstance(value, right_type):
            return Either.right(value)
        else:
            raise TypeError()

    @staticmethod
    def from_optional(value: Optional[B]) -> "Either[None, B]":
        if value is None:
            return Either.left(value)
        return Either.right(value)

    def to_left(self: "Either[AA, NoReturn]") -> "Left[AA]":
        if not isinstance(self, Left):
            raise TypeError("to_left can only be called on an instance of Left.")
        return self

    def to_right(self: "Either[NoReturn, BB]") -> "Right[BB]":
        if not isinstance(self, Right):
            raise TypeError("to_right can only be called on an instance of Right.")
        return self

    def match(
        self,
        case_left: "Callable[[Left[A]], C1]",
        case_right: "Callable[[Right[B]], C2]"
    ) -> Union[C1, C2]:
        if isinstance(self, Left):
            return case_left(self)
        elif isinstance(self, Right):
            return case_right(self)
        else:
            raise TypeError()

    def fold(
        self,
        case_left: "Callable[[A], C1]",
        case_right: "Callable[[B], C2]"
    ) -> Union[C1, C2]:
        return self.match(
            lambda x: case_left(x.value),
            lambda y: case_right(y.value)
        )

    def swap(self) -> "Either[B, A]":
        return self.match(
            lambda left: Either.right(left.value),
            lambda right: Either.left(right.value)
        )

    def map(self, f: Callable[[B], C]) -> "Either[A, C]":
        return self.match(
            lambda left: left,
            lambda right: Either.right(f(right.value))
        )

    def map_left(self, f: Callable[[A], C]) -> "Either[C, B]":
        return self.match(
            lambda left: Either.left(f(left.value)),
            lambda right: right
        )

    def flat_map(self, f: "Callable[[B], Either[AA, C]]") -> "Either[Union[A, AA], C]":
        return self.match(
            lambda left: left,
            lambda right: f(right.value)
        )

    def __lshift__(self, f: "Callable[[B], Either[AA, C]]") -> "Either[Union[A, AA], C]":
        return self.flat_map(f)

    def flatten(self: "Either[A1, Either[AA, BB]]") -> "Either[Union[A1, AA], BB]":
        return self.flat_map(lambda x: x)

    def require(
        self,
        predicate: Callable[[B], bool],
        to_error: Callable[[B], AA]
    ) -> "Either[Union[A, AA], B]":
        def _case_right(right: Right[B]) -> "Either[Union[A, AA], B]":
            if predicate(right.value):
                return right
            return Either.left(to_error(right.value))

        return self.match(
            lambda left: left,
            _case_right
        )

    def asserting(
        self,
        predicate: Callable[[B], bool],
        to_error: Callable[[B], X]
    ) -> "Either[A, B]":
        def _case_right(right: Right[B]) -> "Either[A, B]":
            if not predicate(right.value):
                raise to_error(right.value)
            return right

        return self.match(
            lambda left: left,
            _case_right
        )

    def raise_errors(self: "Either[AA, BB]") -> "Either[NoReturn, BB]":
        def _case_left(error: AA) -> NoReturn:
            if isinstance(error, Exception):
                raise error from error
            else:
                raise EitherException(value=error)

        return self.fold(_case_left, lambda x: Either.right(x))

    def tap(self, op: "Callable[[Either[A, B]], Any]") -> "Either[A, B]":
        op(self)
        return self

    def display(self, description: Optional[str] = None) -> "Either[A, B]":
        if description is not None:
            print(f"{description}:")
        return self.tap(print)

    def to_union(self) -> Union[A, B]:
        return self.match(lambda x: x.value, lambda y: y.value)

    def cast(self: "Either[AA, BB]", t: Type[C]) -> "Either[Union[AA, TypeError], C]":
        return self.flat_map(
            lambda b: Either.right(b) if isinstance(b, t)
            else Either.left(
                TypeError(f"Unable to cast value {b} (of type {type(b).__name__}) as type {t.__name__}.")
            )
        )


@dataclass(frozen=True)
class Left(Generic[A], Either[A, NoReturn]):
    value: A


@dataclass(frozen=True)
class Right(Generic[B], Either[NoReturn, B]):
    value: B


class EitherMonad(Generic[E1]):
    def __lshift__(self, arg: Either[E1, BB]) -> BB:
        return arg.fold(
            lambda e: raise_exception(RaiseLeft(e)),
            lambda a: a
        )


def monadic_method(
    func: Callable[Concatenate[AA, EitherMonad[E], P], Either[E, A]]
) -> Callable[Concatenate[AA, P], Either[E, A]]:
    @functools.wraps(func)
    def _wrapper(self_arg: AA, *args: P.args, **kwargs: P.kwargs) -> Either[E, A]:
        try:
            return func(self_arg, EitherMonad(), *args, **kwargs)
        except RaiseLeft as raise_left:
            # NOTE: WJH (02/03/23) mypy can't prove that the generic type
            #       of the RaiseLeft instance here is `E`, so we have to
            #       use `type: ignore`.
            return Either.left(raise_left.value)  # type: ignore
    return _wrapper


def monadic(
    func: Callable[Concatenate[EitherMonad[E], P], Either[E, A]]
) -> Callable[P, Either[E, A]]:
    @functools.wraps(func)
    def _wrapper(*args: P.args, **kwargs: P.kwargs) -> Either[E, A]:
        try:
            return func(EitherMonad(), *args, **kwargs)
        except RaiseLeft as raise_left:
            # NOTE: WJH (02/03/23) mypy can't prove that the generic type
            #       of the RaiseLeft instance here is `E`, so we have to
            #       use `type: ignore`.
            return Either.left(raise_left.value)  # type: ignore
    return _wrapper


class WithArgs(Protocol[P, A_con, B]):
    def __call__(self, *args: P.args, **kwargs: P.kwargs) -> Callable[[A_con], B]:
        pass


class _Identities:
    def __getitem__(self, t: Type[AA]) -> "EitherArrow[[AA], NoReturn, AA]":
        return EitherArrow.from_callable(lambda x: x)


class EitherArrow(Generic[P, A, B]):
    @staticmethod
    def from_either(either: Either[E, Callable[P, A]]) -> "EitherArrow[P, E, A]":
        return EitherArrow(lambda *args, **kwargs: either.map(lambda f: f(*args, **kwargs)))

    @staticmethod
    def from_callable(f: Callable[P, A]) -> "EitherArrow[P, NoReturn, A]":
        return EitherArrow(lambda *args, **kwargs: Either.right(f(*args, **kwargs)))

    @staticmethod
    def right(x: AA) -> "EitherArrow[[], NoReturn, AA]":
        return EitherArrow(lambda: Either.right(x))

    @staticmethod
    def left(x: BB) -> "EitherArrow[[], BB, NoReturn]":
        return EitherArrow(lambda: Either.left(x))

    identity: Final = _Identities()

    def __init__(self, f: Callable[P, Either[A, B]]) -> None:
        self._f = f

    def to_callable(self) -> "Callable[P, Either[A, B]]":
        return self._f

    def __call__(self, *args: P.args, **kwargs: P.kwargs) -> "Either[A, B]":
        return self._f(*args, **kwargs)

    def catch(
        self: "EitherArrow[P, E, AA]",
        exc: Type[X]
    ) -> "EitherArrow[P, Union[E, X], AA]":
        def _g(*args: P.args, **kwargs: P.kwargs) -> Either[Union[E, X], AA]:
            try:
                return self._f(*args, **kwargs)
            except exc as e:
                return Either.left(e)
        return EitherArrow(_g)

    def to_left(self: "EitherArrow[P, AA, NoReturn]") -> Callable[P, Left[AA]]:
        return lambda *args, **kwargs: self._f(*args, **kwargs).to_left()

    def to_right(self: "EitherArrow[P, NoReturn, BB]") -> Callable[P, Right[BB]]:
        return lambda *args, **kwargs: self._f(*args, **kwargs).to_right()

    def match(
        self,
        case_left: "Callable[[Left[A]], C1]",
        case_right: "Callable[[Right[B]], C2]"
    ) -> Callable[P, Union[C1, C2]]:
        return lambda *args, **kwargs: self._f(*args, **kwargs).match(case_left, case_right)

    def fold(
        self,
        case_left: "Callable[[A], C1]",
        case_right: "Callable[[B], C2]"
    ) -> Callable[P, Union[C1, C2]]:
        return lambda *args, **kwargs: self._f(*args, **kwargs).fold(case_left, case_right)

    def swap(self) -> "EitherArrow[P, B, A]":
        return EitherArrow(lambda *args, **kwargs: self._f(*args, **kwargs).swap())

    def map(self, f: Callable[[B], C]) -> "EitherArrow[P, A, C]":
        return EitherArrow(lambda *args, **kwargs: self._f(*args, **kwargs).map(f))

    def map_left(self, f: Callable[[A], C]) -> "EitherArrow[P, C, B]":
        return EitherArrow(lambda *args, **kwargs: self._f(*args, **kwargs).map_left(f))

    def flat_map(self, f: "Callable[[B], Either[AA, C]]") -> "EitherArrow[P, Union[A, AA], C]":
        return EitherArrow(lambda *args, **kwargs: self._f(*args, **kwargs).flat_map(f))

    def __lshift__(self, f: "Callable[[B], Either[AA, C]]") -> "EitherArrow[P, Union[A, AA], C]":
        return self.flat_map(f)

    def flatten(
        self: "EitherArrow[P, A1, Either[AA, BB]]"
    ) -> "EitherArrow[P, Union[A1, AA], BB]":
        return EitherArrow(lambda *args, **kwargs: self._f(*args, **kwargs).flatten())

    def require(
        self,
        predicate: Callable[[B], bool],
        to_error: Callable[[B], AA]
    ) -> "EitherArrow[P, Union[A, AA], B]":
        return EitherArrow(
            lambda *args, **kwargs: self._f(*args, **kwargs).require(predicate, to_error)
        )

    def asserting(
        self,
        predicate: Callable[[B], bool],
        to_error: Callable[[B], X]
    ) -> "EitherArrow[P, A, B]":
        return EitherArrow(
            lambda *args, **kwargs: self._f(*args, **kwargs).asserting(predicate, to_error)
        )

    def or_else(self, other: "EitherArrow[P, AA, BB]") -> "EitherArrow[P, A | AA, B | BB]":
        return EitherArrow(
            lambda *args, **kwargs: self._f(*args, **kwargs).match(
                lambda _: other(*args, **kwargs),
                lambda a: a
            )
        )

    def raise_errors(self: "EitherArrow[P, AA, BB]") -> "EitherArrow[P, NoReturn, BB]":
        return EitherArrow(lambda *args, **kwargs: self._f(*args, **kwargs).raise_errors())

    def tap(self, op: "Callable[[Either[A, B]], Any]") -> "EitherArrow[P, A, B]":
        return EitherArrow(lambda *args, **kwargs: self._f(*args, **kwargs).tap(op))

    def display(self, description: Optional[str] = None) -> "EitherArrow[P, A, B]":
        return EitherArrow(lambda *args, **kwargs: self._f(*args, **kwargs).display(description))

    def to_union(self) -> Callable[P, Union[A, B]]:
        return lambda *args, **kwargs: self._f(*args, **kwargs).to_union()

    def cast(self: "EitherArrow[P, AA, BB]", t: Type[C]) -> "EitherArrow[P, Union[AA, TypeError], C]":
        return EitherArrow(lambda *args, **kwargs: self._f(*args, **kwargs).cast(t))

    def __getitem__(self: "EitherArrow[P, AA, GetItem[B1, C1]]", key: B1) -> "EitherArrow[P, AA, C1]":
        return EitherArrow(
            lambda *args, **kwargs: self._f(*args, **kwargs).map(lambda x: x[key])
        )

    def match_types(self: "EitherArrow[P, AA, BB]") -> "EitherArrow[P, AA, NoReturn]":
        def _f(arg: BB) -> NoReturn:
            raise TypeMatchException(arg)
        return self.map(_f)

    def at_type(
        self: "EitherArrow[P, AA, BB]",
        target_type: Type[B2],
        operation: "Callable[[B2], Either[A2, C]]"
    ) -> "EitherArrow[P, Union[AA, A2], Union[BB, C]]":
        def _recover(arg: Union[BB, TypeMatchException]) -> Either[A2, Union[BB, C]]:
            if not isinstance(arg, TypeMatchException):
                return Either.right(arg)
            if isinstance(arg.value, target_type):
                return operation(arg.value)
            else:
                raise arg from arg

        return (
            self.swap()
            .catch(TypeMatchException)
            .swap()
            .flat_map(_recover)
        )

    def with_args_map(
        self: "EitherArrow[P, AA, BB]",
        f: WithArgs[P, BB, C]
    ) -> "EitherArrow[P, AA, C]":
        return EitherArrow(lambda *args, **kwargs: self._f(*args, **kwargs).map(lambda x: f(*args, **kwargs)(x)))

    def with_args_flat_map(
        self: "EitherArrow[P, A, BB]",
        f: WithArgs[P, BB, Either[AA, C]]
    ) -> "EitherArrow[P, Union[A, AA], C]":
        return EitherArrow(lambda *args, **kwargs: self._f(*args, **kwargs).flat_map(lambda x: f(*args, **kwargs)(x)))

    def with_args_map_left(
        self: "EitherArrow[P, AA, B]",
        f: WithArgs[P, AA, C]
    ) -> "EitherArrow[P, C, B]":
        return EitherArrow(lambda *args, **kwargs: self._f(*args, **kwargs).map_left(lambda x: f(*args, **kwargs)(x)))



from setuptools import setup, find_namespace_packages


with open("ziopy/VERSION") as version_file:
    version = version_file.read().strip()

kwds = {}
try:
    kwds['long_description'] = open('README.md').read()
    kwds['long_description_content_type'] = 'text/markdown'
except IOError:
    pass

setup(
    name="zio_py",
    version=version,
    author="William Harvey",
    author_email="drwjharvey@gmail.com",
    license='MIT License',
    description="Python port of Scala ZIO for pure functional programming",
    url="https://github.com/miiohio/ziopy",
    packages=["ziopy"] + find_namespace_packages(include=["ziopy.*"]),
    package_data={"ziopy": ["VERSION", "py.typed"]},
    package_dir={"zio_py": "ziopy"},
    install_requires=[
        "dataclasses; python_version<'3.7'",
        "typing_extensions >= 4.2.0",
    ],
    **kwds
)



from dataclasses import dataclass
from typing import Optional, Union


@dataclass(frozen=True)
class Print:
    value: str


@dataclass(frozen=True)
class Input:
    prompt: Optional[str]
    user_input: Union[EOFError, KeyboardInterrupt, str]
